# -*- coding: utf-8 -*-

from __future__ import print_function

import argparse
import os
import re
import subprocess
import sys
if sys.version_info.major == 3:
    from builtins import object
from collections import OrderedDict

from .optionsfile_parser import OptionsFileParser
from .version import __version__


DEFAULT_OPTIONS = {
    "date_format": "%Y-%m-%d",
    "exclude_labels": [],
    "git_remote": "origin",
    "github_api": "api.github.com",
    "github_site": "github.com",
    "header": "# Change Log",
    "issue_prefix": "**Closed issues:**",
    "max_issues": sys.maxsize,
    "merge_prefix": "**Merged pull requests:**",
    "options_file": ".pygcgen",
    "output": "CHANGELOG.md",
    "unreleased_label": "Unreleased",
}


class DebugHelp(argparse.Action):
    # def __init__(self, option_strings, dest, nargs=None, **kwargs):
    #     super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        import platform
        print(self.help)
        print("\n{} v{}".format(parser.prog, __version__))
        print("Python", sys.version)
        print(platform.platform())
        exit()


class OptionsParser(object):
    def __init__(self, options=None):
        self.options = self.parse_options(options)

    def parse_options(self, options):
        parser = argparse.ArgumentParser(
            description='Fully automate changelog generation.',
        )

        parser.add_argument(
            "-u", "--user",
            help="Username of the owner of target GitHub repo"
        )
        parser.add_argument(
            "-p", "--project",
            help="Name of project on GitHub"
        )
        parser.add_argument(
            "-t", "--token",
            help="To make more than 50 requests per hour your GitHub token "
                 "is required. You can generate it at: "
                 "https://github.com/settings/tokens/new"
        )
        parser.add_argument(
            "--options-file", metavar="FILE",
            default=DEFAULT_OPTIONS["options_file"],
            help="Read options from file. Those will overwrite the ones from "
                 "the command line."
        )
        parser.add_argument(
            "-f", "--date-format",
            default=DEFAULT_OPTIONS["date_format"],
            help="The date format to use in changelog. Default is: %%Y-%%m-%%d"
        )
        parser.add_argument(
            "-o", "--output", metavar="FILE",
            default=DEFAULT_OPTIONS["output"],
            help="Output file. Default is CHANGELOG.md"
        )
        parser.add_argument(
            "--no-overwrite", action='store_true',
            help="Don't overwrite the output file if it exists "
                 "(add a number instead)."
        )
        parser.add_argument(
            "-b", "--base", metavar="FILE",
            help="Optional base file to append to generated changelog."
        )
        parser.add_argument(
            "-s", "--section", action="append", nargs="*",
            metavar=('PREFIX', 'LABEL'),
            help="Add a new section to the changelog with the prefix "
                 "'PREFIX'. All issues that match one of the LABEL's "
                 "will be listed in this section."
        )
        lst = [
            ["--header-label", "header", "Setup custom header label."],
            ["--issues-label", "issue_prefix",
             "Setup custom label for closed-issues section."],
            ["--pr-label", "merge_prefix",
             "Setup custom label for pull requests section."]
        ]
        for opt, dest, hlp in lst:
            parser.add_argument(
                opt, dest=dest,
                default=DEFAULT_OPTIONS[dest],
                help="{} Default is: {}".format(hlp, DEFAULT_OPTIONS[dest])
            )
        parser.add_argument(
            "--front-matter", metavar="JSON", dest="frontmatter",
            help="Add YAML front matter. Formatted as JSON because it's "
                 "easier to add on the command line."
        )
        parser.add_argument(
            "--no-issues", action="store_false", dest='issues',
            help="Don't include closed issues in changelog."
        )
        parser.add_argument(
            "--no-issues-wo-labels",
            action="store_true", dest="add_issues_wo_labels",
            help="Don't include closed issues without labels in changelog."
        )
        parser.add_argument(
            "--no-pr-wo-labels",
            action="store_false", dest='add_pr_wo_labels',
            help="Don't include pull requests without labels in changelog."
        )
        parser.add_argument(
            "--no-pull-requests",
            action="store_false", dest='include_pull_request',
            help="Don't include pull-requests in changelog."
        )
        parser.add_argument(
            "--no-filter-by-milestone",
            action="store_false", dest="filter_issues_by_milestone",
            help="Don't use milestone to detect when issue was resolved."
        )
        parser.add_argument(
            "--no-author",
            action="store_false", dest="author",
            help="Don't add author of pull-request in the end."
        )
        parser.add_argument(
            "--author-link-as-tag",
            action='store_true', dest="username_as_tag",
            help="Use GitHub tags instead of Markdown links for the "
                 "author of an issue or pull-request."
        )
        parser.add_argument(
            "--with-unreleased",
            action='store_true', dest="with_unreleased",
            help="Include unreleased closed issues in log."
        )
        parser.add_argument(
            "--unreleased-only",
            action='store_true', dest="unreleased_only",
            help="Generate log from unreleased closed issues only."
        )
        parser.add_argument(
            "--unreleased-label",
            default=DEFAULT_OPTIONS["unreleased_label"],
            help="Label for unreleased closed issues. "
                 "Default is: {0}".format(DEFAULT_OPTIONS["unreleased_label"])
        )
        parser.add_argument(
            "--unreleased-with-date", action='store_true',
            help="Add actual date to unreleased label."
        )
        parser.add_argument(
            "--no-compare-link", action='store_false', dest="compare_link",
            help="Don't include compare link (Full Changelog) between older "
                 "version and newer version."
        )
        parser.add_argument(
            "--include-labels", metavar="LABEL", nargs='*',
            help="Only issues with the specified labels will be "
                 "included in the changelog."
        )
        parser.add_argument(
            "--exclude-labels", metavar="LABEL",
            nargs='*', default=DEFAULT_OPTIONS["exclude_labels"],
            help="Issues with the specified labels will always be "
                 "excluded from changelog. "
                 "Default labels: {0}".format(
                DEFAULT_OPTIONS["exclude_labels"]
            )
        )
        parser.add_argument(
            "--tag-separator",  metavar="SEPARATOR",
            help="The SEPARATOR will be inserted in the log between tags."
        )
        parser.add_argument(
            "--between-tags",  metavar="TAG", nargs='*',
            help="Changelog will be filled only between specified tags."
        )
        parser.add_argument(
            "--exclude-tags", metavar="TAG", nargs='*',
            help="Change log will exclude specified tags."
        )
        parser.add_argument(
            "--exclude-tags-regex",
            help='Apply a regular expression on tag names so that they can be '
            'excluded, for example: --exclude-tags-regex ".*\+\d{1,}"'
        )
        parser.add_argument(
            "--since-tag", metavar="TAG",
            help="Change log will start after specified tag."
        )
        parser.add_argument(
            "--due-tag", metavar="TAG",
            help="Change log will end before specified tag."
        )
        parser.add_argument(
            "--max-issues", metavar="NUMBER",
            type=int, default=DEFAULT_OPTIONS["max_issues"],
            help="Max number of issues to fetch from GitHub. "
                 "Default is unlimited."
        )
        parser.add_argument(
            "--release-url", metavar="URL",
            help="The URL to point to for release links, in printf format "
                 "(with the tag as variable)."
        )
        parser.add_argument(
            "--github-api", metavar="URL",
            dest="github_endpoint", default=DEFAULT_OPTIONS["github_api"],
            help="The enterprise endpoint to use for your Github API."
        )
        parser.add_argument(
            "--github-site", metavar="URL",
            dest="github_site", default=DEFAULT_OPTIONS["github_site"],
            help="The Enterprise Github site on which your project is hosted."
        )
        parser.add_argument(
            "--simple-list", action='store_true',
            help="Create simple list from issues and pull requests. "
        )
        parser.add_argument(
            "--future-release", metavar="RELEASE_VERSION",
            help="Put the unreleased changes in the specified release number."
        )
        parser.add_argument(
            "--release-branch",
            help="Limit pull requests to the release branch, "
                 "such as master or release."
        )
        parser.add_argument(
            "--origin", dest="git_remote",
            default=DEFAULT_OPTIONS["git_remote"],
            help="If you named the origin of your repo other than origin."
        )
        parser.add_argument(
            "-v", "--verbose", action='count', default=False,
            help="Run verbosely."
        )
        parser.add_argument(
            "-q", "--quiet", action='store_true',
            help="Don't output progress information."
        )
        parser.add_argument(
            "--version",
            action='version',
            version="%(prog)s v{0}".format(__version__),
            help="Print version number"
        )
        parser.add_argument(
            "--support", action=DebugHelp, nargs="?",
            help="If you have an issue with pygcgen, got to "
                 "https://github.com/topic2k/pygcgen/issues and add a "
                 "new issue. \nAdd this information to your issue discription."
        )
        opts = parser.parse_args(options)

        if os.path.exists(opts.options_file):
            OptionsFileParser(options=opts).parse()
        if not opts.user or not opts.project:
            self.fetch_user_and_project(opts)

        sections = OrderedDict()
        if opts.section:
            for s in opts.section:
                labels = []
                for l in s[1:]:
                    # this is to remove empty label strings (could happen if
                    # in the config file a section line ends with a comma.)
                    if l:
                        labels.append(l)
                sections.update({s[0]: labels})

        opts.sections = sections
        del opts.section

        return opts

    def fetch_user_and_project(self, options):
        user, project = self.user_and_project_from_git(options)
        if not options.user:
            options.user = user
        if not options.project:
            options.project = project

    def user_and_project_from_git(self, options, arg0=None, arg1=None):
        """ Detects user and project from git. """
        user, project = self.user_project_from_option(options, arg0, arg1)
        if user and project:
            return user, project

        try:
            remote = subprocess.check_output(
                [
                    'git', 'config', '--get',
                    'remote.{0}.url'.format(options.git_remote)
                ]
            )
        except subprocess.CalledProcessError:
            return None, None
        except WindowsError:
            print("git binary not found.")
            exit(1)
        else:
            return self.user_project_from_remote(remote)

    @staticmethod
    def user_project_from_option(options, arg0, arg1):
        """
        Try to find user and project name from git remote output

        @param [String] output of git remote command
        @return [Array] user and project
        """

        site = options.github_site
        if arg0 and not arg1:
            # this match should parse strings such as
            #   "https://github.com/skywinder/Github-Changelog-Generator"
            # or
            #   "skywinder/Github-Changelog-Generator"
            #  to user and project
            match = re.match(
                "(?:.+{site}/)?(.+)/(.+)".format(site=site),
                arg0
            )
            if not match:
                print("Can't detect user and name from first "
                      "parameter: '{arg0}' -> exit'".format(arg0=arg0))
                exit(1)
            return match.groups()
        return None, None

    @staticmethod
    def user_project_from_remote(remote):
        """
        Try to find user and project name from git remote output

        @param [String] output of git remote command
        @return [Array] user and project
        """

        # try to find repo in format:
        # origin	git@github.com:skywinder/Github-Changelog-Generator.git (fetch)
        # git@github.com:skywinder/Github-Changelog-Generator.git
        regex1 = br".*(?:[:/])(?P<user>(-|\w|\.)*)/" \
                 br"(?P<project>(-|\w|\.)*)(\.git).*"
        match = re.match(regex1, remote)
        if match:
            return match.group("user"), match.group("project")

        # try to find repo in format:
        # origin	https://github.com/skywinder/ChangelogMerger (fetch)
        # https://github.com/skywinder/ChangelogMerger
        regex2 = r".*/((?:-|\w|\.)*)/((?:-|\w|\.)*).*"
        match = re.match(regex2, remote)
        if match:
            return match.group("user"), match.group("project")

        return None, None



# -*- coding: utf-8 -*-

from __future__ import print_function
import os

from pygcgen import ChangelogGenerator


base_options = [
    "--quiet",
    # "-h",
    # "-v",
    # "-vv",  # or "-v", "-v",
    # "-vvv",
    # "--options-file", ".pygcgen_example",
    # "-u", "topic2k",
    # "-p", "pygcgen",
    # '-s', "**Questions:**", "question", "Question",
    # '-s', "**Future-Requests:**", "feature-request",
    # '--section', '**Important changes:**', 'notice',
    # '-s', "**Implemented enhancements:**", "enhancement", "Enhancement",
    # '-s', "**Fixed bugs:**", "bug", "Bug",
    # "-v",
    # "--no-overwrite",
    # "--between-tags", "v0.1.1",
    # "--include-labels", "bug",
    # "--no-issues-wo-labels",
    # "--future-release", "v0.2.0",
    # "--tag-separator", " ---\n\n",
]

on_travis = os.environ.get('TRAVIS', 'false') == 'true'
if not on_travis:
    ChangelogGenerator(base_options + ["-v"]).run()
else:
    tests = [
        [  # Test #01
            "--no-overwrite",
            "--section", '**Important changes:**', 'notice',
            "--since-tag", "v0.1.0",
            "--between-tags", "v0.1.1", "v0.2.1",
            "--due-tag", "v0.2.0",
            "--exclude-tags-regex", "v0\.0\..*",
            "--exclude-tags", "v0.1.2",
            "--with-unreleased",
            "--include-labels", "notice", "enhancement", "bug",
            "--exclude-labels",
            "duplicate", "Duplicate",
            "invalid", "Invalid",
            "wontfix", "Wontfix",
            "question", "Question",
            "hide in changelog",
        ]
    ]
    for nr, options in enumerate(tests, start=1):
        print("starting test {} ...".format(nr), end="")
        ChangelogGenerator(base_options + options).run()
        print(" done.")



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import os
import re
import sys
import subprocess
import threading
if sys.version_info.major == 3:
    from builtins import object, range

from agithub.GitHub import GitHub

from .pygcgen_exceptions import GithubApiError


GH_CFG_VARS = ["github.pygcgen.token", "github.token"]
PER_PAGE_NUMBER = 100
MAX_SIMULTANEOUS_REQUESTS = 25
CHANGELOG_GITHUB_TOKEN = "CHANGELOG_GITHUB_TOKEN"
GH_RATE_LIMIT_EXCEEDED_MSG = \
    "GitHub API rate limit exceeded, change log may be missing some issues. " \
    "Please provide a token with -t option or in git config."
NO_TOKEN_PROVIDED = \
    "Warning: No token provided. Neither -t option, git config or variable " \
    "$CHANGELOG_GITHUB_TOKEN found. This script can make only " \
    "50 requests to GitHub API per hour without token!"
REPO_CREATED_TAG_NAME = "repo_created_at"


class Fetcher(object):
    """
    A Fetcher is responsible for all requests to GitHub and all basic
    manipulation with related data (such as filtering, validating, e.t.c).
    """

    def __init__(self, options):
        self.options = options
        self.first_issue = None
        self.events_cnt = 0
        self.fetch_github_token()
        if isinstance(self.options.user, bytes):
            self.options.user = self.options.user.decode("utf8")
        if isinstance(self.options.project, bytes):
            self.options.project = self.options.project.decode("utf8")
        if isinstance(self.options.token, bytes):
            self.options.token = self.options.token.decode("utf8")
        if options.token:
            self.github = GitHub(
                token=options.token,
                api_url=options.github_endpoint
            )
        else:
            self.github = GitHub(api_url=options.github_endpoint)

    def fetch_github_token(self):
        """
        Fetch GitHub token. First try to use variable provided
        by --token option, otherwise try to fetch it from git config
        and last CHANGELOG_GITHUB_TOKEN env variable.

        :returns: Nothing
        """

        if not self.options.token:
            try:
                for v in GH_CFG_VARS:
                    cmd = ['git', 'config', '--get', '{0}'.format(v)]
                    self.options.token = subprocess.Popen(
                        cmd, stdout=subprocess.PIPE).communicate()[0].strip()
                    if self.options.token:
                        break
            except subprocess.CalledProcessError:
                pass
        if not self.options.token:
            self.options.token = os.environ.get(CHANGELOG_GITHUB_TOKEN)
        if not self.options.token:
            print(NO_TOKEN_PROVIDED)

    def get_all_tags(self):
        """
        Fetch all tags for repository from Github.

        :return: tags in repository
        :rtype: list
        """

        verbose = self.options.verbose
        gh = self.github
        user = self.options.user
        repo = self.options.project
        if verbose:
            print("Fetching tags...")

        tags = []
        page = 1
        while page > 0:
            if verbose > 2:
                print(".", end="")
            rc, data = gh.repos[user][repo].tags.get(
                page=page, per_page=PER_PAGE_NUMBER)
            if rc == 200:
                tags.extend(data)
            else:
                self.raise_GitHubError(rc, data, gh.getheaders())
            page = NextPage(gh)
        if verbose > 2:
            print(".")

        if len(tags) == 0:
            if not self.options.quiet:
                print("Warning: Can't find any tags in repo. Make sure, that "
                      "you push tags to remote repo via 'git push --tags'")
                exit()
        if verbose > 1:
            print("Found {} tag(s)".format(len(tags)))
        return tags

    def fetch_closed_issues_and_pr(self):
        """
        This method fetches all closed issues and separate them to
        pull requests and pure issues (pull request is kind of issue
        in term of GitHub).

        :rtype: list, list
        :return: issues, pull-requests
        """

        verbose = self.options.verbose
        gh = self.github
        user = self.options.user
        repo = self.options.project
        if verbose:
            print("Fetching closed issues and pull requests...")

        data = []
        issues = []
        data = []
        page = 1
        while page > 0:
            if verbose > 2:
                print(".", end="")
            rc, data = gh.repos[user][repo].issues.get(
                page=page, per_page=PER_PAGE_NUMBER,
                state='closed', filter='all'
            )
            if rc == 200:
                issues.extend(data)
            else:
                self.raise_GitHubError(rc, data, gh.getheaders())
            if len(issues) >= self.options.max_issues:
                break
            page = NextPage(gh)
        self.first_issue = data[-1] if len(data) > 0 else []
        if verbose > 2:
            print(".")

        # separate arrays of issues and pull requests:
        prs = []
        iss = []
        for i in issues:
            if "pull_request" in i:
                prs.append(i)
            else:
                iss.append(i)
        if verbose > 1:
            print("\treceived {} issues and  {} pull requests.".format(
                len(iss), len(prs))
            )
        return iss, prs

    def fetch_closed_pull_requests(self):
        """
        Fetch all pull requests. We need them to detect "merged_at" parameter

        :rtype: list
        :return: all pull requests
        """

        pull_requests = []
        verbose = self.options.verbose
        gh = self.github
        user = self.options.user
        repo = self.options.project
        if verbose:
            print("Fetching closed pull requests...")
        page = 1
        while page > 0:
            if verbose > 2:
                print(".", end="")

            if self.options.release_branch:
                rc, data = gh.repos[user][repo].pulls.get(
                    page=page, per_page=PER_PAGE_NUMBER, state='closed',
                    base=self.options.release_branch
                )
            else:
                rc, data = gh.repos[user][repo].pulls.get(
                    page=page, per_page=PER_PAGE_NUMBER, state='closed',
                )

            if rc == 200:
                pull_requests.extend(data)
            else:
                self.raise_GitHubError(rc, data, gh.getheaders())
            page = NextPage(gh)
        if verbose > 2:
            print(".")
        if verbose > 1:
            print("\tfetched {} closed pull requests.".format(
                len(pull_requests))
            )
        return pull_requests

    def fetch_repo_creation_date(self):
        """
        Get the creation date of the repository from GitHub.

        :rtype: str, str
        :return: special tag name, creation date as ISO date string
        """
        gh = self.github
        user = self.options.user
        repo = self.options.project
        rc, data = gh.repos[user][repo].get()
        if rc == 200:
            return REPO_CREATED_TAG_NAME, data["created_at"]
        else:
            self.raise_GitHubError(rc, data, gh.getheaders())
        return None, None

    def fetch_events_async(self, issues, tag_name):
        """
        Fetch events for all issues and add them to self.events

        :param list issues: all issues
        :param str tag_name: name of the tag to fetch events for
        :returns: Nothing
        """

        if not issues:
            return issues
        verbose = self.options.verbose
        gh = self.github
        user = self.options.user
        repo = self.options.project
        self.events_cnt = 0
        if verbose:
            print("fetching events for {} {}... ".format(
                len(issues), tag_name)
            )

        def worker(issue):
            page = 1
            issue['events'] = []
            while page > 0:
                rc, data = gh.repos[user][repo].issues[
                    issue['number']].events.get(
                    page=page, per_page=PER_PAGE_NUMBER)
                if rc == 200:
                    issue['events'].extend(data)
                    self.events_cnt += len(data)
                else:
                    self.raise_GitHubError(rc, data, gh.getheaders())
                page = NextPage(gh)

        threads = []
        cnt = len(issues)
        for i in range(0, (cnt // MAX_SIMULTANEOUS_REQUESTS) + 1):
            for j in range(MAX_SIMULTANEOUS_REQUESTS):
                idx = i * MAX_SIMULTANEOUS_REQUESTS + j
                if idx == cnt:
                    break
                t = threading.Thread(target=worker, args=(issues[idx],))
                threads.append(t)
                t.start()
                if verbose > 2:
                    print(".", end="")
                    if not idx % PER_PAGE_NUMBER:
                        print("")
            for t in threads:
                t.join()
        if verbose > 2:
            print(".")

    def fetch_date_of_tag(self, tag):
        """
        Fetch time for tag from repository.

        :param dict tag: dictionary with tag information
        :rtype: str
        :return: time of specified tag as ISO date string
        """

        if self.options.verbose > 1:
            print("\tFetching date for tag {}".format(tag["name"]))
        gh = self.github
        user = self.options.user
        repo = self.options.project

        rc, data = gh.repos[user][repo].git.commits[
            tag["commit"]["sha"]].get()
        if rc == 200:
            return data["committer"]["date"]
        self.raise_GitHubError(rc, data, gh.getheaders())

    def fetch_commit(self, event):
        """
        Fetch commit data for specified event.

        :param dict event: dictionary with event information
        :rtype: dict
        :return: dictionary with commit data
        """

        gh = self.github
        user = self.options.user
        repo = self.options.project

        rc, data = gh.repos[user][repo].git.commits[
            event["commit_id"]].get()
        if rc == 200:
            return data
        self.raise_GitHubError(rc, data, gh.getheaders())

    @staticmethod
    def raise_GitHubError(rc, data, header):
        hdr = dict(header)
        if rc == 403 and hdr.get("x-ratelimit-remaining") == '0':
            # TODO: add auto-retry
            raise GithubApiError(GH_RATE_LIMIT_EXCEEDED_MSG)
        raise GithubApiError("({0}) {1}".format(rc, data["message"]))


def NextPage(gh):
    """
    Checks if a GitHub call returned multiple pages of data.

    :param gh: GitHub() instance
    :rtype: int
    :return: number of next page or 0 if no next page
    """
    header = dict(gh.getheaders())
    if 'Link' in header:
        parts = header['Link'].split(',')
        for part in parts:
            subparts = part.split(';')
            sub = subparts[1].split('=')
            if sub[0].strip() == 'rel':
                if sub[1] == '"next"':
                    page = int(
                        re.match(
                            r'.*page=(\d+).*', subparts[0],
                            re.IGNORECASE | re.DOTALL | re.UNICODE
                        ).groups()[0]
                    )
                    return page
    return 0



# -*- coding: utf-8 -*-

from __future__ import print_function

import os
import re
import sys
if sys.version_info.major == 3:
    from builtins import object

from .generator import Generator
from .options_parser import OptionsParser
from .pygcgen_exceptions import ChangelogGeneratorError


def checkname(filename):
    if not os.path.exists(filename):
        return filename
    mtch = re.match(r'(?P<filename>.*?)(?P<nr>\d+)(?P<rest>($|.md)?)',
                    filename)
    if mtch:
        # filename exists, add a number
        nr = int(mtch.group("nr")) + 1
        filename = "{0}{1}{2}".format(mtch.group("filename"), nr,
                                      mtch.group("rest"))
    else:
        # filename exists, but doesn't end with a number -> add one.
        filename = filename.rsplit(".", 1)
        filename = filename[0] + "_1" + (("." + filename[1]) if
                                         len(filename) > 1 else "")
    return checkname(filename)


class ChangelogGenerator(object):
    """ Class responsible for whole change log generation cycle. """

    def __init__(self, options=None):
        """
        :type options: list
        :param options: command line arguments
        """

        self.options = OptionsParser(options).options
        self.generator = Generator(self.options)

    def run(self):
        """
        The entry point of this script to generate change log
        'ChangelogGeneratorError' Is thrown when one
        of the specified tags was not found in list of tags.
        """
        if not self.options.project or not self.options.user:
            print("Project and/or user missing. "
                  "For help run:\n  pygcgen --help")
            return

        if not self.options.quiet:
            print("Generating changelog...")

        log = None
        try:
            log = self.generator.compound_changelog()
        except ChangelogGeneratorError as err:
            print("\n\033[91m\033[1m{}\x1b[0m".format(err.args[0]))
            exit(1)
        if not log:
            if not self.options.quiet:
                print("Empty changelog generated. {} not written.".format(
                    self.options.output)
                )
            return

        if self.options.no_overwrite:
            out = checkname(self.options.output)
        else:
            out = self.options.output

        with open(out, "w") as fh:
            try:
                fh.write(log.encode("utf8"))
            except TypeError:
                fh.write(log)
        if not self.options.quiet:
            print("Done!")
            print("Generated changelog written to {}".format(out))


def run():
    ChangelogGenerator(sys.argv[1:]).run()


# def run_gui():
#     import wx
#     from .gui import GeneratorApp
#
#
#     app = GeneratorApp()
#     wx.MessageBox("Not yet implemented.", os.path.basename(sys.argv[0]))
#     # win = MainFrame(None).Show()
#     # app.MainLoop()
#


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-

import re
import sys
if sys.version_info.major == 3:
    from builtins import object


# Pattern to split list options (respecting quotes)
regex = re.compile(
    '\s*(^|,)\s*(["\'])?(?(2)(?P<label1>.*?)\\2|(?P<label2>.*?)((?=,)|$))'
)

FILENAME = ".pygcgen"
KNOWN_INTEGER_KEYS = ["max_issues"]
KNOWN_ARRAY_KEYS = [
    "between_tags",
    "exclude_labels",
    "exclude_tags"
    "include_labels",
    "section",
]
KNOWN_MULTIPLE_OPTIONS = {
    # False means the option wasn't used in config file
    # Used to overwrite value given by commandline
    "section": False
}
IRREGULAR_OPTIONS = {
    "front_matter": "frontmatter",
    "github_api": "github_endpoint",
    "header_label": "header",
    "issues_label": "issue_prefix",
    "no_author": "author",
    "no_compare_link": "compare_link",
    "no_filter_by_milestone": "filter_issues_by_milestone",
    "no_issues": "issues",
    "no_issues_wo_labels": "add_issues_wo_labels",
    "no_pr_wo_labels": "add_pr_wo_labels",
    "no_pull_requests": "include_pull_request",
    "origin": "git_remote",
    "pr_label": "merge_prefix",
    "author_link_as_tag": "username_as_tag",
}
BOOL_KEYS = {
    "debug": True,
    "no_author": False,
    "no_compare_link": False,
    "no_filter_by_milestone": False,
    "no_issues": False,
    "no_issues_wo_labels": True,
    "no_overwrite": True,
    "no_pr_wo_labels": False,
    "no_pull_requests": False,
    "quiet": False,
    "simple_list": True,
    "unreleased_only": True,
    "unreleased_with_date": True,
    "username_as_tag": False,
    "verbose": True,
    "with_unreleased": True,
}


class ParserError(Exception):
    pass


# ParserFile is a configuration file reader which sets options in the
# given Hash.
#
# In your project's root, you can put a file named
# <tt>.pygcgen</tt> to override defaults.
#
# Example:
#   header_label=# My Super Changelog
#   ; Comments are allowed
#   future-release=5.0.0
#   # Ruby-style comments, too
#   since-tag=1.0.0
#
# The configuration format is
#    <tt>some-key=value</tt>
# or
#    <tt>some_key=value</tt>.
#

class OptionsFileParser(object):
    def __init__(self, options):
        # @param options [Hash] options to be configured from file contents
        # @param file [nil,IO] configuration file handle, defaults to
        # opening `.pygcgen`
        self.options = options
        with open(options.options_file, "r") as fh:
            self.filecontents = fh.read().split("\n")

    def parse(self):
        # Sets options using configuration file content
        for line in self.filecontents:
            self.parse_line(line)
        return self.options

    def parse_line(self, line):
        if not line or self.non_configuration_line(line):
            return
        option_name, value = self.extract_pair(line)

        if option_name in IRREGULAR_OPTIONS:
            option_name = IRREGULAR_OPTIONS[option_name]

        if option_name in KNOWN_MULTIPLE_OPTIONS.keys():
            if self.options.__dict__[option_name] and \
                KNOWN_MULTIPLE_OPTIONS[option_name]:
                self.options.__dict__[option_name].append(
                    self.convert_value(value, option_name)
                )
            else:
                self.options.__dict__[option_name] = [
                    self.convert_value(value, option_name),
                ]
                KNOWN_MULTIPLE_OPTIONS[option_name] = True
        else:
            self.options.__dict__[option_name] = self.convert_value(
                value, option_name
            )

    @staticmethod
    def non_configuration_line(line):
        # Returns true if the line starts with a pound sign or a semi-colon.
        return line.startswith(';') or line.startswith('#')

    @staticmethod
    def extract_pair(line):
        # Returns the option name as a symbol and its string value.
        #
        # @param line [String] unparsed line from config file
        # @return [Array<Symbol, String>]
        try:
            key, value = line.split("=", 2)
        except ValueError:
            key = line
            value = ""
        key = key.strip().lower().replace("-", "_")
        value = value.replace("\\n", "\n")
        return key, value

    @staticmethod
    def convert_value(value, option_name):
        if option_name in KNOWN_ARRAY_KEYS:
            val = []
            for item in regex.finditer(value):
                val.append(
                    item.group("label1") if item.group("label1")
                    else item.group("label2")
                )
            value = val
        elif option_name in KNOWN_INTEGER_KEYS:
            value = int(value)
        elif option_name in BOOL_KEYS:
            if not value:
                value = BOOL_KEYS[option_name]
            elif value.isdigit():
                value = bool(value)
            elif not value.isdigit():
                value = value.strip()[0].lower() in ["t", "y" "1"]
        return value



# -*- coding: utf-8 -*-

__version__ = "0.2.2"

__title__ = "pygcgen"
__summary__ = "Automatic changelog generation"
__uri__ = "https://github.com/topic2k/pygcgen"

__author__ = "topic2k"
__email__ = "topic2k+pypi@gmail.com"

__license__ = "MIT"
__copyright__ = "2016 %s" % __author__



# -*- coding: utf-8 -*-

import wx


class MainFrame(wx.Frame):
    def __init__(self, *args, **kwargs):
        super(MainFrame, self).__init__(*args, **kwargs)
        self.BuildGui()

    def BuildGui(self):
        pass


class GeneratorApp(wx.App):
    pass



# -*- coding: utf-8 -*-

###########  original file info from reader.rb  ####################
### #######  https://github.com/skywinder/github-changelog-generator
#
# Author:: Enrico Stahn <mail@enricostahn.com>
#
# Copyright 2014, Zanui, <engineering@zanui.com.au>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
####################################################################

# A Reader to read an existing ChangeLog file and return a structured object
#
# Example:
#   reader = GitHubChangelogGenerator::Reader.new
#   content = reader.read('./CHANGELOG.md')

import re
import sys
if sys.version_info.major == 3:
    from builtins import map


def parse_heading(heading):
    """
    Parse a single heading and return a Hash
    The following heading structures are currently valid:
    - ## [v1.0.2](https://github.com/zanui/chef-thumbor/tree/v1.0.1) (2015-03-24)
    - ## [v1.0.2](https://github.com/zanui/chef-thumbor/tree/v1.0.1)
    - ## v1.0.2 (2015-03-24)
    - ## v1.0.2

    @param [String] heading Heading from the ChangeLog File
    @return [Hash] Returns a structured Hash with version, url and date
    """

    heading_structures = [
        r"^## \[(?P<version>.+?)\]\((?P<url>.+?)\)( \((?P<date>.+?)\))?$",
        r"^## (?P<version>.+?)( \((?P<date>.+?)\))?$",
        ]
    captures = {"version": None, "url": None, "date": None}

    for regexp in heading_structures:
        matches = re.match(regexp, heading)
        if matches:
            captures.update(matches.groupdict())
            break
    return captures


def parse(data):
    """
    Parse the given ChangeLog data into a list of Hashes.

    @param [String] data File data from the ChangeLog.md
    @return [Array<Hash>] Parsed data, e.g. [{ 'version' => ..., 'url' => ..., 'date' => ..., 'content' => ...}, ...]
    """

    sections = re.compile("^## .+$", re.MULTILINE).split(data)
    headings = re.findall("^## .+?$", data, re.MULTILINE)
    sections.pop(0)
    parsed = []

    def func(h, s):
        p = parse_heading(h)
        p["content"] = s
        parsed.append(p)

    list(map(func, headings, sections))
    return parsed


def read_changelog(options):
    with open(options.base, 'r') as fh:
        log = fh.read()
        log = log.lstrip(options.header).strip()
        return parse(log)



# -*- coding: utf-8 -*-

from setuptools import setup

version = {}
with open("pygcgen/version.py") as fh:
    exec(fh.read(), version)

setup(
    name=version['__title__'],
    version=version['__version__'],
    license=version['__license__'],

    keywords=version['__summary__'],
    description='Generate changelog based on tags, issues and '
                'merged pull requests on GitHub.',
    long_description="This package started as a conversion from "
    "ruby to python of the 'GitHub Changelog Generator' "
    "(https://github.com/skywinder/github-changelog-generator).",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development :: Documentation",
    ],

    author=version['__author__'],
    author_email=version['__email__'],
    maintainer=version['__author__'],
    maintainer_email=version['__email__'],

    url=version['__uri__'],

    install_requires=["agithub", "python-dateutil"],

    packages=[version['__title__']],
    # data_files=[
    #     ('..', [
    #         'README.md',
    #         'LICENSE',
    #         'CHANGELOG.md',
    #         '.pygcgen_example'
    #     ]),
    # ],
    exclude_package_data={
        '.': ['.gitignore', ]
    },

    entry_points={
        'console_scripts': ['pygcgen = pygcgen.main:run', ],
        #'gui_scripts': ['pygcgenw = pygcgen.main:run_gui', ],
    },
)



# -*- coding: utf-8 -*-

from __future__ import division, print_function

import copy
import datetime
import re
import sys
import threading
if sys.version_info.major == 3:
    from builtins import object, range, str
from collections import OrderedDict

import dateutil.tz
from dateutil.parser import parse as dateutil_parser

from .fetcher import Fetcher, REPO_CREATED_TAG_NAME
from .pygcgen_exceptions import ChangelogGeneratorError
from .reader import read_changelog


def timestring_to_datetime(timestring):
    """
    Convert an ISO formated date and time string to a datetime object.

    :param str timestring: String with date and time in ISO format.
    :rtype: datetime
    :return: datetime object
    """
    result = dateutil_parser(str(timestring))
    return result


# noinspection PyTypeChecker
class Generator(object):
    """
    A Generator responsible for all logic, related with
    change log generation from ready-to-parse issues.
    """

    def __init__(self, options):
        self.options = options
        self.tag_times_dict = {}
        self.issues = []
        self.issues2 = []
        self.pull_requests = []
        self.all_tags = []
        self.filtered_tags = []
        self.fetcher = Fetcher(options)

    def fetch_and_filter_issues_and_pr(self):
        issues, pull_requests = self.fetcher.fetch_closed_issues_and_pr()

        if self.options.verbose:
            print("Filtering issues and pull requests...")
        self.pull_requests = []
        if self.options.include_pull_request:
            self.pull_requests = self.get_filtered_pull_requests(pull_requests)

        self.issues = []
        if self.options.issues:
            self.issues = self.filter_by_labels(issues, "issues")

        self.fetch_events_for_issues_and_pr()
        self.issues = self.detect_actual_closed_dates(self.issues, "issues")
        self.pull_requests = self.detect_actual_closed_dates(
            self.pull_requests, "pull requests"
        )

    def fetch_events_for_issues_and_pr(self):
        """
        Fetch event for issues and pull requests

        @return [Array] array of fetched issues
        """

        # Async fetching events:
        self.fetcher.fetch_events_async(self.issues, "issues")
        self.fetcher.fetch_events_async(self.pull_requests, "pull requests")

    def fetch_tags_dates(self):
        """ Async fetching of all tags dates. """

        if self.options.verbose:
            print("Fetching dates for {} tags...".format(len(self.filtered_tags)))

        def worker(tag):
            self.get_time_of_tag(tag)

        # Async fetching tags:
        threads = []
        max_threads = 50
        cnt = len(self.filtered_tags)
        for i in range(0, (cnt // max_threads) + 1):
            for j in range(max_threads):
                idx = i * 50 + j
                if idx == cnt:
                    break
                t = threading.Thread(target=worker,
                                     args=(self.filtered_tags[idx],))
                threads.append(t)
                t.start()
                if self.options.verbose > 2:
                    print(".", end="")
            for t in threads:
                t.join()
        if self.options.verbose > 2:
            print(".")
        if self.options.verbose > 1:
            print("Fetched dates for {} tags.".format(
                len(self.tag_times_dict))
            )

    def detect_actual_closed_dates(self, issues, kind):
        """
        Find correct closed dates, if issues was closed by commits.

        :param list issues: issues to check
        :param str kind: either "issues" or "pull requests"
        :rtype: list
        :return: issues with updated closed dates
        """

        if self.options.verbose:
            print("Fetching closed dates for {} {}...".format(
                len(issues), kind)
            )
        all_issues = copy.deepcopy(issues)
        for issue in all_issues:
            if self.options.verbose > 2:
                print(".", end="")
                if not issues.index(issue) % 30:
                    print("")
            self.find_closed_date_by_commit(issue)

            if not issue.get('actual_date', False):
                if issue.get('closed_at', False):
                    print("Skipping closed non-merged issue: #{0} {1}".format(
                        issue["number"], issue["title"]))

                all_issues.remove(issue)

        if self.options.verbose > 2:
            print(".")
        return all_issues

    def find_closed_date_by_commit(self, issue):
        """
        Fill "actual_date" parameter of specified issue by closed date of
        the commit, if it was closed by commit.

        :param dict issue: issue to edit
        """

        if not issue.get('events'):
            return
        # if it's PR -> then find "merged event", in case
        # of usual issue -> find closed date
        compare_string = "merged" if 'merged_at' in issue else "closed"
        # reverse! - to find latest closed event. (event goes in date order)
        # if it were reopened and closed again.
        issue['events'].reverse()
        found_date = False
        for event in issue['events']:
            if event["event"] == compare_string:
                self.set_date_from_event(event, issue)
                found_date = True
                break
        if not found_date:
            # TODO: assert issues, that remain without
            #       'actual_date' hash for some reason.
            print("\nWARNING: Issue without 'actual_date':"
                  " #{0} {1}".format(issue["number"], issue["title"]))

    def set_date_from_event(self, event, issue):
        """
        Set closed date from this issue.

        :param dict event: event data
        :param dict issue: issue data
        """

        if not event.get('commit_id', None):
            issue['actual_date'] = timestring_to_datetime(issue['closed_at'])
            return
        try:
            commit = self.fetcher.fetch_commit(event)
            issue['actual_date'] = timestring_to_datetime(
                commit['author']['date']
            )
        except ValueError:
            print("WARNING: Can't fetch commit {0}. "
                  "It is probably referenced from another repo.".
                  format(event['commit_id']))
            issue['actual_date'] = timestring_to_datetime(issue['closed_at'])

    @staticmethod
    def encapsulate_string(raw_string):
        """
        Encapsulate characters to make markdown look as expected.

        :param str raw_string: string to encapsulate
        :rtype: str
        :return: encapsulated input string
        """

        raw_string.replace('\\', '\\\\')
        enc_string = re.sub("([<>*_()\[\]#])", r"\\\1", raw_string)
        return enc_string

    def compound_changelog(self):
        """
        Main function to start change log generation

        :rtype: str
        :return: Generated change log file
        """

        self.fetch_and_filter_tags()
        tags_sorted = self.sort_tags_by_date(self.filtered_tags)
        self.filtered_tags = tags_sorted
        self.fetch_and_filter_issues_and_pr()

        log = str(self.options.frontmatter) \
            if self.options.frontmatter else u""
        log += u"{0}\n\n".format(self.options.header)

        if self.options.unreleased_only:
            log += self.generate_unreleased_section()
        else:
            log += self.generate_log_for_all_tags()

        try:
            with open(self.options.base) as fh:
                log += fh.read()
        except (TypeError, IOError):
            pass
        return log

    def generate_sub_section(self, issues, prefix):
        """
        Generate formated list of issues for changelog.

        :param list issues: Issues to put in sub-section.
        :param str prefix: Title of sub-section.
        :rtype: str
        :return: Generated ready-to-add sub-section.
        """

        log = ""
        if issues:
            if not self.options.simple_list:
                log += u"{0}\n\n".format(prefix)
            for issue in issues:
                merge_string = self.get_string_for_issue(issue)
                log += u"- {0}\n".format(merge_string)
            log += "\n"
        return log

    def generate_header(self, newer_tag_name, newer_tag_link,
                        newer_tag_time,
                        older_tag_link, project_url):
        """
        Generate a header for a tag section with specific parameters.

        :param str newer_tag_name: Name (title) of newer tag.
        :param str newer_tag_link: Tag name of newer tag, used for links.
                               Could be same as **newer_tag_name** or some
                               specific value, like `HEAD`.
        :param datetime newer_tag_time: Date and time when
                                        newer tag was created.
        :param str older_tag_link: Tag name of older tag, used for links.
        :param str project_url: URL for current project.
        :rtype: str
        :return: Generated ready-to-add tag section.
        """

        log = ""
        # Generate date string:
        # noinspection PyUnresolvedReferences
        time_string = newer_tag_time.strftime(self.options.date_format)

        # Generate tag name and link
        if self.options.release_url:
            release_url = self.options.release_url.format(newer_tag_link)
        else:
            release_url = u"{project_url}/tree/{newer_tag_link}".format(
                project_url=project_url, newer_tag_link=newer_tag_link)

        if not self.options.unreleased_with_date and \
                newer_tag_name == self.options.unreleased_label:
            log += u"# [{newer_tag_name}]({release_url})\n\n".format(
                newer_tag_name=newer_tag_name, release_url=release_url)
        else:
            log += u"# [{newer_tag_name}]({release_url}) " \
                   u"({time_string})\n".format(
                        newer_tag_name=newer_tag_name,
                        release_url=release_url,
                        time_string=time_string
                   )

        if self.options.compare_link \
            and older_tag_link != REPO_CREATED_TAG_NAME:
            # Generate compare link
            log += u"[Full Changelog]"
            log += u"({project_url}/compare/{older_tag_link}".format(
                project_url=project_url,
                older_tag_link=older_tag_link,
            )
            log += u"...{newer_tag_link})\n\n".format(
                newer_tag_link=newer_tag_link
            )
        return log

    def generate_log_between_tags(self, older_tag, newer_tag):
        """
        Generate log between 2 specified tags.

        :param dict older_tag: All issues before this tag's date will be
                               excluded. May be special value, if new tag is
                               the first tag. (Means **older_tag** is when
                               the repo was created.)
        :param dict newer_tag: All issues after this tag's date  will be
                               excluded. May be title of unreleased section.
        :rtype: str
        :return: Generated ready-to-add tag section for newer tag.
        """

        filtered_issues, filtered_pull_requests = \
            self.filter_issues_for_tags(newer_tag, older_tag)

        older_tag_name = older_tag["name"] if older_tag \
            else self.detect_since_tag()

        if not filtered_issues and not filtered_pull_requests:
            # do not generate an unreleased section if it would be empty
            return ""
        return self.generate_log_for_tag(
            filtered_pull_requests, filtered_issues,
            newer_tag, older_tag_name)

    def filter_issues_for_tags(self, newer_tag, older_tag):
        """
        Apply all filters to issues and pull requests.

        :param dict older_tag: All issues before this tag's date will be
                               excluded. May be special value, if new tag is
                               the first tag. (Means **older_tag** is when
                               the repo  was created.)
        :param dict newer_tag: All issues after this tag's date  will be
                               excluded. May be title of unreleased section.
        :rtype: list(dict), list(dict)
        :return: Filtered issues and pull requests.
        """

        filtered_pull_requests = self.delete_by_time(self.pull_requests,
                                                     older_tag, newer_tag)
        filtered_issues = self.delete_by_time(self.issues, older_tag,
                                              newer_tag)

        newer_tag_name = newer_tag["name"] if newer_tag else None

        if self.options.filter_issues_by_milestone:
            # delete excess irrelevant issues (according milestones).Issue #22.
            filtered_issues = self.filter_by_milestone(
                filtered_issues, newer_tag_name, self.issues
            )
            filtered_pull_requests = self.filter_by_milestone(
                filtered_pull_requests, newer_tag_name, self.pull_requests
            )
        return filtered_issues, filtered_pull_requests

    def generate_log_for_all_tags(self):
        """
        The full cycle of generation for whole project.

        :rtype: str
        :return: The complete change log for released tags.
        """

        if self.options.verbose:
            print("Generating log...")
        self.issues2 = copy.deepcopy(self.issues)

        log1 = ""
        if self.options.with_unreleased:
            log1 = self.generate_unreleased_section()

        log = ""
        for index in range(len(self.filtered_tags) - 1):
            log += self.do_generate_log_for_all_tags_part1(log, index)

        if self.options.tag_separator and log1:
            log = log1 + self.options.tag_separator + log
        else:
            log = log1 + log

        if len(self.filtered_tags) != 0:
            log += self.do_generate_log_for_all_tags_part2(log)

        return log

    def do_generate_log_for_all_tags_part1(self, log, index):
        if self.options.verbose > 1:
            print("\tgenerate log for {}".format(
                self.filtered_tags[index]["name"]))
        log2 = self.generate_log_between_tags(
            self.filtered_tags[index + 1], self.filtered_tags[index])
        if self.options.tag_separator and log and log2:
            return self.options.tag_separator + log2
        return log2

    def do_generate_log_for_all_tags_part2(self, log):
        older_tag = self.last_older_tag()
        if self.options.verbose > 1:
            print("\tgenerate log for {}".format(
                self.filtered_tags[-1]["name"]))
        log2 = self.generate_log_between_tags(
            older_tag, self.filtered_tags[-1])
        if self.options.tag_separator and log and log2:
            return self.options.tag_separator + log2
        return log2

    def last_older_tag(self):
        older_tag = {"name": self.get_temp_tag_for_repo_creation()}
        if self.options.between_tags or self.options.since_tag:
            older_tag_date = self.get_time_of_tag(older_tag)
            newer_tag_date = self.get_time_of_tag(self.filtered_tags[-1])
            for tag in self.all_tags:
                tag_date = self.get_time_of_tag(tag)
                if older_tag_date < tag_date < newer_tag_date:
                    older_tag = tag
                    older_tag_date = tag_date
        return older_tag

    def generate_unreleased_section(self):
        """
        Generate log for unreleased closed issues.

        :rtype: str
        :return: Generated ready-to-add unreleased section.
        """
        if not self.filtered_tags:
            return ""
        now = datetime.datetime.utcnow()
        now = now.replace(tzinfo=dateutil.tz.tzutc())
        head_tag = {"name": self.options.unreleased_label}
        self.tag_times_dict[head_tag["name"]] = now
        unreleased_log = self.generate_log_between_tags(
            self.filtered_tags[0], head_tag)
        return unreleased_log

    def get_string_for_issue(self, issue):
        """
        Parse issue and generate single line formatted issue line.

        Example output:
            - Add coveralls integration [\#223](https://github.com/skywinder/github-changelog-generator/pull/223) ([skywinder](https://github.com/skywinder))
            - Add coveralls integration [\#223](https://github.com/skywinder/github-changelog-generator/pull/223) (@skywinder)


        :param dict issue: Fetched issue from GitHub.
        :rtype: str
        :return: Markdown-formatted single issue.
        """

        encapsulated_title = self.encapsulate_string(issue['title'])
        try:
            title_with_number = u"{0} [\\#{1}]({2})".format(
                encapsulated_title, issue["number"], issue["html_url"]
            )
        except UnicodeEncodeError:
            # TODO: why did i add this? Is it needed?
            title_with_number = "ERROR ERROR ERROR: #{0} {1}".format(
                issue["number"], issue['title']
            )
            print(title_with_number, '\n', issue["html_url"])
        return self.issue_line_with_user(title_with_number, issue)

    def issue_line_with_user(self, line, issue):
        """
        If option author is enabled, a link to the profile of the author
        of the pull reqest will be added to the issue line.

        :param str line: String containing a markdown-formatted single issue.
        :param dict issue: Fetched issue from GitHub.
        :rtype: str
        :return: Issue line with added author link.
        """
        if not issue.get("pull_request") or not self.options.author:
            return line

        if not issue.get("user"):
            line += u" (Null user)"
        elif self.options.username_as_tag:
            line += u" (@{0})".format(
                issue["user"]["login"]
            )
        else:
            line += u" ([{0}]({1}))".format(
                issue["user"]["login"], issue["user"]["html_url"]
            )
        return line

    def generate_log_for_tag(self,
                             pull_requests,
                             issues,
                             newer_tag,
                             older_tag_name):
        """
        Generates log for tag section with header and body.

        :param list(dict) pull_requests: List of PR's in this tag section.
        :param list(dict) issues: List of issues in this tag section.
        :param dict newer_tag: Github data of tag for this section.
        :param str older_tag_name: Older tag, used for the links.
                                   May be special value, if **newer tag** is
                                   the first tag. (Means **older_tag** is when
                                   the repo was created.)
        :rtype: str
        :return: Ready-to-add and parsed tag section.
        """

        newer_tag_link, newer_tag_name, \
        newer_tag_time = self.detect_link_tag_time(newer_tag)

        github_site = "https://github.com" or self.options.github_endpoint
        project_url = "{0}/{1}/{2}".format(
            github_site, self.options.user, self.options.project)

        log = self.generate_header(newer_tag_name, newer_tag_link,
                                   newer_tag_time, older_tag_name, project_url)
        if self.options.issues:
            # Generate issues:
            log += self.issues_to_log(issues, pull_requests)
        if self.options.include_pull_request:
            # Generate pull requests:
            log += self.generate_sub_section(
                pull_requests, self.options.merge_prefix
            )
        return log

    def issues_to_log(self, issues, pull_requests):
        """
        Generate ready-to-paste log from list of issues and pull requests.

        :param list(dict) issues: List of issues in this tag section.
        :param list(dict) pull_requests: List of PR's in this tag section.
        :rtype: str
        :return: Generated log for issues and pull requests.
        """

        log = ""
        sections_a, issues_a = self.parse_by_sections(
            issues, pull_requests)

        for section, s_issues in sections_a.items():
            log += self.generate_sub_section(s_issues, section)
        log += self.generate_sub_section(issues_a, self.options.issue_prefix)
        return log

    def parse_by_sections(self, issues, pull_requests):
        """
        This method sort issues by types (bugs, features, etc. or
        just closed issues) by labels.

        :param list(dict) issues: List of issues in this tag section.
        :param list(dict) pull_requests: List of PR's in this tag section.
        :rtype: dict(list(dict)), list(dict)
        :return: Issues and PR's sorted into sections.
        """

        issues_a = []
        sections_a = OrderedDict()

        if not self.options.sections:
            return [sections_a, issues]
        for key in self.options.sections:
            sections_a.update({key: []})
        self.parse_by_sections_for_issues(issues, sections_a, issues_a)
        self.parse_by_sections_for_pr(pull_requests, sections_a)
        return [sections_a, issues_a]

    def parse_by_sections_for_issues(self, issues, sections_a, issues_a):
        for section, sect_labels in self.options.sections.items():
            added_issues = []
            for issue in issues:
                is_labels = issue.get('labels')
                if is_labels:
                    is_lbls = set(l["name"] for l in is_labels)
                    if is_lbls.intersection(set(sect_labels)):
                        sections_a[section].append(issue)
                        added_issues.append(issue)
                        continue
                issues_a.append(issue)
                added_issues.append(issue)
            for iss in added_issues:
                issues.remove(iss)

    def parse_by_sections_for_pr(self, pull_requests, sections_a):
        for section, sect_labels in self.options.sections.items():
            added_pull_requests = []
            for pr in pull_requests:
                pr_labels = pr.get('labels')
                if pr_labels:
                    pr_lbls = set(l["name"] for l in pr_labels)
                    if pr_lbls.intersection(set(sect_labels)):
                        sections_a[section].append(pr)
                        added_pull_requests.append(pr)
                        continue
            for pr in added_pull_requests:
                pull_requests.remove(pr)

    def exclude_issues_by_labels(self, issues):
        """
        Delete all issues with labels from exclude-labels option.

        :param list(dict) issues: All issues for tag.
        :rtype: list(dict)
        :return: Filtered issues.
        """
        if not self.options.exclude_labels:
            return copy.deepcopy(issues)

        remove_issues = set()
        exclude_labels = self.options.exclude_labels
        include_issues = []
        for issue in issues:
            for label in issue["labels"]:
                if label["name"] in exclude_labels:
                    remove_issues.add(issue["number"])
                    break
        for issue in issues:
            if issue["number"] not in remove_issues:
                include_issues.append(issue)
        return include_issues

    def filter_by_milestone(self, filtered_issues, tag_name, all_issues):
        """
        :param list(dict) filtered_issues: Filtered issues.
        :param str tag_name: Name (title) of tag.
        :param list(dict) all_issues: All issues.
        :rtype: list(dict)
        :return: Filtered issues according milestone.
        """

        filtered_issues = self.remove_issues_in_milestones(filtered_issues)
        if tag_name:
            # add missed issues (according milestones)
            issues_to_add = self.find_issues_to_add(all_issues, tag_name)
            filtered_issues.extend(issues_to_add)
        return filtered_issues

    @staticmethod
    def find_issues_to_add(all_issues, tag_name):
        """
        Add all issues, that should be in that tag, according to milestone.

        :param list(dict) all_issues: All issues.
        :param str tag_name: Name (title) of tag.
        :rtype: List[dict]
        :return: Issues filtered by milestone.
        """

        filtered = []
        for issue in all_issues:
            if issue.get("milestone"):
                if issue["milestone"]["title"] == tag_name:
                    iss = copy.deepcopy(issue)
                    filtered.append(iss)
        return filtered

    def remove_issues_in_milestones(self, filtered_issues):
        """
        :param list(dict) filtered_issues: Filtered issues.
        :rtype: list(dict)
        :return: List with removed issues, that contain milestones with
                 same name as a tag.
        """
        for issue in filtered_issues:
            # leave issues without milestones
            if issue["milestone"]:
                # check, that this milestone is in tag list:
                for tag in self.filtered_tags:
                    if tag["name"] == issue["milestone"]["title"]:
                        filtered_issues.remove(issue)
        return filtered_issues

    def delete_by_time(self, issues, older_tag, newer_tag):
        """
        Filter issues that belong to specified tag range.

        :param list(dict) issues: Issues to filter.
        :param dict older_tag: All issues before this tag's date will be
                               excluded. May be special value, if **newer_tag**
                               is the first tag. (Means **older_tag** is when
                               the repo was created.)
        :param dict newer_tag: All issues after this tag's date  will be
                               excluded. May be title of unreleased section.
        :rtype: list(dict)
        :return: Filtered issues.
        """

        if not older_tag and not newer_tag:
            # in case if no tags are specified - return unchanged array
            return copy.deepcopy(issues)

        newer_tag_time = self.get_time_of_tag(newer_tag)
        older_tag_time = self.get_time_of_tag(older_tag)
        filtered = []
        for issue in issues:
            if issue.get('actual_date'):
                rslt = older_tag_time < issue['actual_date'] <= newer_tag_time
                if rslt:
                    filtered.append(copy.deepcopy(issue))
        return filtered

    def include_issues_by_labels(self, all_issues):
        """
        Include issues with labels, specified in self.options.include_labels.

        :param list(dict) all_issues: All issues.
        :rtype: list(dict)
        :return: Filtered issues.
        """

        included_by_labels = self.filter_by_include_labels(all_issues)
        wo_labels = self.filter_wo_labels(all_issues)
        il = set([f["number"] for f in included_by_labels])
        wl = set([w["number"] for w in wo_labels])
        filtered_issues = []
        for issue in all_issues:
            if issue["number"] in il or issue["number"] in wl:
                filtered_issues.append(issue)
        return filtered_issues

    def filter_wo_labels(self, all_issues):
        """
        Filter all issues that don't have a label.

        :rtype: list(dict)
        :return: Issues without labels.
        """

        issues_wo_labels = []
        if not self.options.add_issues_wo_labels:
            for issue in all_issues:
                if not issue['labels']:
                    issues_wo_labels.append(issue)
        return issues_wo_labels

    def filter_by_include_labels(self, issues):
        """
        Filter issues to include only issues with labels
        specified in include_labels.

        :param list(dict) issues: Pre-filtered issues.
        :rtype: list(dict)
        :return: Filtered issues.
        """

        if not self.options.include_labels:
            return copy.deepcopy(issues)
        filtered_issues = []
        include_labels = set(self.options.include_labels)
        for issue in issues:
            labels = [label["name"] for label in issue["labels"]]
            if include_labels.intersection(labels):
                filtered_issues.append(issue)
        return filtered_issues

    def filter_by_labels(self, all_issues, kind):
        """
        Filter issues for include/exclude labels.

        :param list(dict) all_issues: All issues.
        :param str kind: Either "issues" or "pull requests".
        :rtype: list(dict)
        :return: Filtered issues.
        """

        filtered_issues = self.include_issues_by_labels(all_issues)
        filtered = self.exclude_issues_by_labels(filtered_issues)
        if self.options.verbose > 1:
            print("\tremaining {}: {}".format(kind, len(filtered)))
        return filtered

    def get_filtered_pull_requests(self, pull_requests):
        """
        This method fetches missing params for PR and filter them
        by specified options. It include add all PR's with labels
        from options.include_labels and exclude all from
        options.exclude_labels.

        :param list(dict) pull_requests: All pull requests.
        :rtype: list(dict)
        :return: Filtered pull requests.
        """

        pull_requests = self.filter_by_labels(pull_requests, "pull requests")
        pull_requests = self.filter_merged_pull_requests(pull_requests)
        if self.options.verbose > 1:
            print("\tremaining pull requests: {}".format(len(pull_requests)))
        return pull_requests

    def filter_merged_pull_requests(self, pull_requests):
        """
        This method filter only merged PR and fetch missing required
        attributes for pull requests. Using merged date is more correct
        than closed date.

        :param list(dict) pull_requests: Pre-filtered pull requests.
        :rtype: list(dict)
        :return:
        """

        if self.options.verbose:
            print("Fetching merge date for pull requests...")
        closed_pull_requests = self.fetcher.fetch_closed_pull_requests()

        if not pull_requests:
            return []
        pulls = copy.deepcopy(pull_requests)
        for pr in pulls:
            fetched_pr = None
            for fpr in closed_pull_requests:
                if fpr['number'] == pr['number']:
                    fetched_pr = fpr
            if fetched_pr:
                pr['merged_at'] = fetched_pr['merged_at']
                closed_pull_requests.remove(fetched_pr)

        for pr in pulls:
            if not pr.get('merged_at'):
                pulls.remove(pr)
        return pulls

    def fetch_and_filter_tags(self):
        """
        Fetch and filter tags, fetch dates and sort them in time order.
        """

        self.all_tags = self.fetcher.get_all_tags()
        self.filtered_tags = self.get_filtered_tags(self.all_tags)
        self.fetch_tags_dates()

    def sort_tags_by_date(self, tags):
        """
        Sort all tags by date.

        :param list(dict) tags: All tags.
        :rtype: list(dict)
        :return: Sorted list of tags.
        """

        if self.options.verbose:
            print("Sorting tags...")
        tags.sort(key=lambda x: self.get_time_of_tag(x))
        tags.reverse()
        return tags

    def get_time_of_tag(self, tag):
        """
        Get date and time for tag, fetching it if not already cached.

        :param dict tag: Tag to get the datetime for.
        :rtype: datetime
        :return: datetime for specified tag.
        """

        if not tag:
            raise ChangelogGeneratorError("tag is nil")

        name_of_tag = tag["name"]
        time_for_name = self.tag_times_dict.get(name_of_tag, None)
        if time_for_name:
            return time_for_name
        else:
            time_string = self.fetcher.fetch_date_of_tag(tag)
            try:
                self.tag_times_dict[name_of_tag] = \
                    timestring_to_datetime(time_string)
            except UnicodeWarning:
                print("ERROR ERROR:", tag)
                self.tag_times_dict[name_of_tag] = \
                    timestring_to_datetime(time_string)
            return self.tag_times_dict[name_of_tag]

    def detect_link_tag_time(self, tag):
        """
        Detect link, name and time for specified tag.

        :param dict tag: Tag data.
        :rtype: str, str, datetime
        :return: Link, name and time of the tag.
        """

        # if tag is nil - set current time
        newer_tag_time = self.get_time_of_tag(tag) if tag \
            else datetime.datetime.now()

        # if it's future release tag - set this value
        if tag["name"] == self.options.unreleased_label \
            and self.options.future_release:
            newer_tag_name = self.options.future_release
            newer_tag_link = self.options.future_release
        elif tag["name"] is not self.options.unreleased_label :
            # put unreleased label if there is no name for the tag
            newer_tag_name = tag["name"]
            newer_tag_link = newer_tag_name
        else:
            newer_tag_name = self.options.unreleased_label
            newer_tag_link = "HEAD"
        return [newer_tag_link, newer_tag_name, newer_tag_time]

    def detect_since_tag(self):
        """
        Try to find tag name to use as older tag for range of log creation.

        :rtype: str
        :return: Tag name to use as 'oldest' tag. May be special value,
                 indicating the creation of the repo.
        """
        return self.options.since_tag or self.version_of_first_item()

    def version_of_first_item(self):
        """
        Try to detect the newest tag from self.options.base, otherwise
        return a special value indicating the creation of the repo.

        :rtype: str
        :return: Tag name to use as 'oldest' tag. May be special value,
                 indicating the creation of the repo.
        """
        try:
            sections = read_changelog(self.options)
            return sections[0]["version"]
        except(IOError, TypeError):
            return self.get_temp_tag_for_repo_creation()

    def get_temp_tag_for_repo_creation(self):
        """
        If not already cached, fetch the creation date of the repo, cache it
        and return the special value indicating the creation of the repo.

        :rtype: str
        :return: value indicating the creation
        """
        tag_date = self.tag_times_dict.get(REPO_CREATED_TAG_NAME, None)
        if not tag_date:
            tag_name, tag_date = self.fetcher.fetch_repo_creation_date()
            self.tag_times_dict[tag_name] = timestring_to_datetime(tag_date)
        return REPO_CREATED_TAG_NAME

    def get_filtered_tags(self, all_tags):
        """
        Return tags after filtering tags in lists provided by
        option: --between-tags & --exclude-tags

        :param list(dict) all_tags: All tags.
        :rtype: list(dict)
        :return: Filtered tags.
        """

        filtered_tags = self.filter_since_tag(all_tags)
        if self.options.between_tags:
            filtered_tags = self.filter_between_tags(filtered_tags)
        if self.options.due_tag:
            filtered_tags = self.filter_due_tag(filtered_tags)
        return self.filter_excluded_tags(filtered_tags)

    def filter_since_tag(self, all_tags):
        """
        Filter tags according since_tag option.

        :param list(dict) all_tags: All tags.
        :rtype: list(dict)
        :return: Filtered tags.
        """

        tag = self.detect_since_tag()
        if not tag or tag == REPO_CREATED_TAG_NAME:
            return copy.deepcopy(all_tags)

        filtered_tags = []
        tag_names = [t["name"] for t in all_tags]
        try:
            idx = tag_names.index(tag)
        except ValueError:
            self.warn_if_tag_not_found(tag, "since-tag")
            return copy.deepcopy(all_tags)

        since_tag = all_tags[idx]
        since_date = self.get_time_of_tag(since_tag)
        for t in all_tags:
            tag_date = self.get_time_of_tag(t)
            if since_date <= tag_date:
                filtered_tags.append(t)
        return filtered_tags

    def filter_due_tag(self, all_tags):
        """
        Filter tags according due_tag option.

        :param list(dict) all_tags: Pre-filtered tags.
        :rtype: list(dict)
        :return: Filtered tags.
        """

        filtered_tags = []
        tag = self.options.due_tag
        tag_names = [t["name"] for t in all_tags]
        try:
            idx = tag_names.index(tag)
        except ValueError:
            self.warn_if_tag_not_found(tag, "due-tag")
            return copy.deepcopy(all_tags)

        due_tag = all_tags[idx]
        due_date = self.get_time_of_tag(due_tag)
        for t in all_tags:
            tag_date = self.get_time_of_tag(t)
            if tag_date <= due_date:
                filtered_tags.append(t)
        return filtered_tags

    def filter_between_tags(self, all_tags):
        """
        Filter tags according between_tags option.

        :param list(dict) all_tags: Pre-filtered tags.
        :rtype: list(dict)
        :return: Filtered tags.
        """

        tag_names = [t["name"] for t in all_tags]
        between_tags = []
        for tag in self.options.between_tags:
            try:
                idx = tag_names.index(tag)
            except ValueError:
                raise ChangelogGeneratorError(
                    "ERROR: can't find tag {0}, specified with "
                    "--between-tags option.".format(tag))
            between_tags.append(all_tags[idx])

        between_tags = self.sort_tags_by_date(between_tags)

        if len(between_tags) == 1:
            # if option --between-tags was only 1 tag given, duplicate it
            # to generate the changelog only for that one tag.
            between_tags.append(between_tags[0])

        older = self.get_time_of_tag(between_tags[1])
        newer = self.get_time_of_tag(between_tags[0])

        for tag in all_tags:
            if older < self.get_time_of_tag(tag) < newer:
                between_tags.append(tag)
        if older == newer:
            between_tags.pop(0)
        return between_tags

    def filter_excluded_tags(self, all_tags):
        """
        Filter tags according exclude_tags and exclude_tags_regex option.

        :param list(dict) all_tags: Pre-filtered tags.
        :rtype: list(dict)
        :return: Filtered tags.
        """
        filtered_tags = copy.deepcopy(all_tags)
        if self.options.exclude_tags:
            filtered_tags = self.apply_exclude_tags(filtered_tags)
        if self.options.exclude_tags_regex:
            filtered_tags = self.apply_exclude_tags_regex(filtered_tags)
        return filtered_tags

    def apply_exclude_tags_regex(self, all_tags):
        """
        Filter tags according exclude_tags_regex option.

        :param list(dict) all_tags: Pre-filtered tags.
        :rtype: list(dict)
        :return: Filtered tags.
        """
        filtered = []
        for tag in all_tags:
            if not re.match(self.options.exclude_tags_regex, tag["name"]):
                filtered.append(tag)
        if len(all_tags) == len(filtered):
            self.warn_if_nonmatching_regex()
        return filtered

    def apply_exclude_tags(self, all_tags):
        """
        Filter tags according exclude_tags option.

        :param list(dict) all_tags: Pre-filtered tags.
        :rtype: list(dict)
        :return: Filtered tags.
        """
        filtered = copy.deepcopy(all_tags)
        for tag in all_tags:
            if tag["name"] not in self.options.exclude_tags:
                self.warn_if_tag_not_found(tag, "exclude-tags")
            else:
                filtered.remove(tag)
        return filtered

    def warn_if_nonmatching_regex(self):
        if not self.options.quiet:
            print(
                "WARNING: unable to reject any tag, using regex "
                "'{0}' in --exclude-tags-regex option.".format(
                    self.options.exclude_tags_regex
                )
            )

    def warn_if_tag_not_found(self, tag, option):
        if not self.options.quiet:
            print(
                "WARNING: can't find tag '{0}' specified with "
                "--{1} option.".format(tag, option)
            )
